import { defineQuery, hasComponent } from 'bitecs'
import { query as q } from './query-parser/index.js'
import { parseAll } from './parser.js'
import { curry, of, is, map, when, assoc, isNil, path, pipe, always, assocPath, prop } from './fn.js'
import { Relationship } from './query-parser/types.js'

const prepareQuery = (query, component) => {
  const { match, returnValues } = query
  const relationship = is(Relationship, match)

  if (relationship) {
    const { from: f, to: t, edge: e } = match
    const [from, to, edge] = map(q => prepareQuery({ match: q, returnValues }, component), [f, t, e])

    return { from, to, edge, relationship, schema: query }
  } else {
    const label = match.label.value
    const name = match.name.value
    const type = component[match.label.value]

    return { name, label, type, relationship, query: defineQuery([type]), schema: query }
  }
}

const matches = (world, type, query, set = new Set()) => entity => {
  if (set.has(entity)) {
    return true
  } else if (hasComponent(world, type, entity) && query.includes(entity)) {
    set.add(entity)
    return true
  } else {
    return false
  }
}

const queryRelationship = (query, world) => {
  const [from, edge, to] = ['from', 'edge', 'to'].map(name => {
    let node = query[name]
    node = node.relationship ? node.from : node
    return { ...node, results: node.query(world) }
  })

  const matchesFrom = matches(world, from.type, from.results)
  const matchesTo = matches(world, to.type, to.results)

  const Edge = edge.type

  const result = []
  for (let i = 0; i < edge.results.length; i++) {
    const eid = edge.results[i]

    if (matchesFrom(Edge.from[eid]) && matchesTo(Edge.to[eid])) {
      result.push(eid)
    }
  }

  return result
}

const assembleQuery = (query, entities) => map(entity => ([{ entity, query }]), entities)

const executeQuery = curry((query, world) => {
  if (!query.relationship) {
    return assembleQuery(query, query.query(world))
  }

  const edges = queryRelationship(query, world)

  if (query.to.relationship) {
    return edges.reduce((acc, eid) => {
      const next = assocPath(
        'from.query',
        always(of(query.edge.type.to[eid])),
        query.to
      )

      return executeQuery(next, world)
        .map(child => [eid, ...child])
        .map(([l, r]) => [{ entity: l, query }, r])
        .concat(acc)
    }, [])
  } else {
    return assembleQuery(query, edges)
  }
})

const isTag = type => (
  Object.getOwnPropertySymbols(type).find(
    (s) => s.description === 'tagStore'
  )
)

const getValue = (type, entity) => {
  if (isTag(type)) {
    return {}
  } else {
    return Object.fromEntries(
      Object.entries(type).map(([k, v]) => [k, v[entity]])
    )
  }
}

const resolveNode = curry((node, obj) => {
  const name = path('query.schema.match.name', node)
  const returns = path('query.schema.returnValues', node)
  const { entity, query: { type } } = node
  const value = getValue(type, entity)
  return returns.includes(name) ? { ...obj, [name.value]: value } : obj
})

const resolveRelationship = edges => {
  const { entity, query } = edges
  const Edge = query.edge.type

  const to = query.to?.from ?? query.to
  return pipe(
    resolveNode({ entity: Edge.from[entity], query: query.from }),
    resolveNode({ entity, query: query.edge }),
    resolveNode({ entity: Edge.to[entity], query: to }),
  )({})
}

const resolveReturns = query => {
  return query.map(edges => {
    const returns = edges.map(edge => {
      if (edge.query.relationship) {
        return resolveRelationship(edge)
      } else {
        return resolveNode(edge, {})
      }
    })

    return Object.assign({}, ...returns)
  })
}

export const query = curry((input, { world, component }) => {
  return parseAll(q, input).map(({ use, ...rest }) => {
    const useWorld = world[use?.value ?? 'default']
    const preparedQuery = prepareQuery(rest, component)
    const results = executeQuery(preparedQuery, useWorld)
    const returns = resolveReturns(results)
    return returns
  })
})

